<?php

namespace Drupal\sdc_component_library\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Theme\ComponentPluginManager;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Finder\Finder;

/**
 * Provides route responses for component previews.
 */
class ComponentsController extends ControllerBase {

  /**
   * The Component Plugin Manager service.
   *
   * @var \Drupal\Core\Theme\ComponentPluginManager
   */
  private ComponentPluginManager $pluginManager;

  /**
   * The path to the components directory in the active theme.
   *
   * @var string
   */
  private string $componentsPath;

  /**
   * Constructs a ComponentsController object.
   *
   * @param string $componentsPath
   *   The path to the components directory in the active theme.
   * @param \Drupal\Core\Theme\ComponentPluginManager $pluginManager
   *   The component plugin manager service from Drupal Core.
   */
  public function __construct(
    string $componentsPath,
    ComponentPluginManager $pluginManager,
  ) {
    $this->componentsPath = $componentsPath;
    $this->pluginManager = $pluginManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    $themeManager = $container->get('theme.manager');
    $activeTheme = $themeManager->getActiveTheme();
    $themePath = $container->get('extension.list.theme')->getPath($activeTheme->getName());

    return new static(
      $themePath . '/components',
      $container->get('plugin.manager.sdc')
    );
  }

  /**
   * Provides a page with a list of component previews.
   *
   * @return array
   *   A render array containing the component previews.
   */
  public function content(): array {
    try {
      // Get all available components from the plugin manager.
      $allComponents = $this->pluginManager->getDefinitions();

      // Find components, including those missing .story.twig files.
      $components = $this->findComponentsWithWarnings($allComponents);

      return $this->buildRenderArray($components);
    }
    catch (\RuntimeException $e) {
      return [
        '#markup' => $this->t('Error: @message', ['@message' => $e->getMessage()]),
      ];
    }
  }

  /**
   * Find components and provide a warning if .story.twig is missing.
   *
   * @param array $allComponents
   *   An array of all components.
   *
   * @return array
   *   An array of components data with warnings if necessary.
   */
  private function findComponentsWithWarnings(array $allComponents): array {
    $components = [];

    foreach ($allComponents as $componentDefinition) {
      $componentName = $componentDefinition['machineName'];

      // Search for the .story.twig file.
      $storyTwigFile = $this->findStoryTwigFile($componentName);

      // Add component data with or without template file.
      $components[] = $storyTwigFile
        ? $this->buildComponentData($storyTwigFile, $this->getRelativePathFromComponents($storyTwigFile), $componentDefinition)
        : [
          'title' => $componentDefinition['name'],
          'provider' => $componentDefinition['provider'] ?? '',
          'description' => 'This component is missing a .story.twig file. Please create one to enable previews.',
          'type' => $componentName,
          'twig_template' => NULL,
          'missing_template' => TRUE,
        ];
    }

    return $components;
  }

  /**
   * Search for the .story.twig file for a component.
   *
   * @param string $componentName
   *   The machine name of the component.
   *
   * @return string|null
   *   The full path to the .story.twig file if found, otherwise null.
   */
  private function findStoryTwigFile(string $componentName): ?string {
    $finder = new Finder();
    $finder->files()
      ->in($this->componentsPath)
      ->name($componentName . '.story.twig');

    foreach ($finder as $file) {
      return $file->getRealPath();
    }

    return NULL;
  }

  /**
   * Get the relative path from the "components/" directory.
   *
   * @param string $fullPath
   *   The full file path of the component.
   *
   * @return string
   *   The relative path starting from "components/".
   */
  private function getRelativePathFromComponents(string $fullPath): string {
    return strpos($fullPath, '/components/') !== FALSE
      ? substr($fullPath, strpos($fullPath, '/components/') + 1)
      : $fullPath;
  }

  /**
   * Build component data from file paths.
   *
   * @param string $realPath
   *   The real file path of the component.
   * @param string $relativePath
   *   The relative file path of the component.
   * @param array $componentDefinition
   *   The component definition data.
   *
   * @return array
   *   The component data array.
   */
  private function buildComponentData(string $realPath, string $relativePath, array $componentDefinition): array {
    return $this->getDefaultComponentData($relativePath, $componentDefinition);
  }

  /**
   * Get default component data structure.
   *
   * @param string $relativePath
   *   The relative file path of the component.
   * @param array $componentDefinition
   *   The component definition data.
   *
   *   Provides a page with a list of component previews.
   *
   * @return array
   *   A render array containing the component previews.
   */
  private function getDefaultComponentData(string $relativePath, array $componentDefinition): array {
    return [
      'title' => $componentDefinition['name'] ?? 'Unnamed Component',
      'provider' => $componentDefinition['provider'] ?? '',
      'description' => $componentDefinition['description'] ?? '',
      'group' => $componentDefinition['group'] ?? '',
      'props' => $componentDefinition['props'] ?? [],
      'example_data' => $componentDefinition['example_data'] ?? [],
      'libraryDependencies' => $componentDefinition['libraryDependencies'] ?? [],
      'type' => explode('/', $relativePath)[0],
      'theme' => 'component_preview_item',
      'twig_template' => '@' . $componentDefinition['provider'] . '/' . $relativePath,
    ];
  }

  /**
   * Build the render array for component previews.
   *
   * @param array $components
   *   An array of component data.
   *
   * @return array
   *   The render array.
   */
  private function buildRenderArray(array $components): array {
    return [
      '#theme' => 'component_preview',
      '#components' => $components,
      '#attached' => [
        'library' => [
          'sdc_component_library/axe_core',
          'sdc_component_library/components_preview',
        ],
      ],
    ];
  }

}
